# process_all_v2.py
# An improved script that uses a dedicated 'audios' folder for intermediate files
# and cleans them up upon successful completion.

import gradio as gr
import torch
import os
import sys
import subprocess
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed
import traceback
import time

# funasr imports
from funasr import AutoModel
from funasr.utils.postprocess_utils import rich_transcription_postprocess

# --- Part 1: Logic from toMp3.py (Modified for dedicated output folder) ---

def check_ffmpeg():
    """Checks if ffmpeg is installed and available in the system's PATH."""
    try:
        subprocess.run(['ffmpeg', '-version'], check=True, capture_output=True)
        return True
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False

def convert_to_mp3(video_path: Path):
    """
    Converts a single video file to an MP3, saving it in the 'audios' subfolder.
    Returns the output path on success, or an error string on failure.
    """
    # Force output to be in the 'audios' subfolder
    output_path = Path("audios") / video_path.with_suffix('.mp3').name
    video_filename = video_path.name
    
    command = [
        'ffmpeg', '-i', str(video_path), '-q:a', '0', '-map', 'a',
        '-y', '-hide_banner', '-loglevel', 'error', str(output_path)
    ]
    try:
        subprocess.run(command, check=True, capture_output=True, text=True)
        # Return the Path object on success
        return output_path
    except subprocess.CalledProcessError as e:
        # Return an error string on failure
        return f"❌ Error converting '{video_filename}': {e.stderr.strip()}"
    except Exception as e:
        return f"❌ An unexpected error occurred with '{video_filename}': {e}"

# --- Part 2: Model Initialization (Same as before) ---
# Create necessary folders at the start
os.makedirs("audios", exist_ok=True)
os.makedirs("sub", exist_ok=True)

print("正在加载模型，请稍候...")
device = "cuda:0" if torch.cuda.is_available() else "cpu"
model = AutoModel(
    model="iic/SenseVoiceSmall",
    trust_remote_code=True,
    vad_model="fsmn-vad",
    vad_kwargs={"max_single_segment_time": 30000},
    device=device,
)
print("模型加载完成。")


# --- Part 3: The Main Orchestrator Function (NEW and IMPROVED) ---

def video_to_subtitle_pipeline(folder_path, language):
    """
    The main pipeline function that:
    1. Converts MP4s to MP3s in a dedicated 'audios' folder.
    2. Transcribes all audio (original and converted) to SRTs.
    3. Cleans up the intermediate audio files upon success.
    """
    # --- STAGE 1: VIDEO TO AUDIO CONVERSION ---
    yield "--- STAGE 1 of 3: Converting Videos to Audio ---"
    
    if not folder_path or not os.path.isdir(folder_path):
        yield "Error: Please provide a valid folder path."
        return

    if not check_ffmpeg():
        yield "FATAL ERROR: ffmpeg is not found. Cannot convert videos. Please install and add to PATH."
        return

    folder = Path(folder_path)
    mp4_files = list(folder.glob('*.mp4'))
    created_audio_paths = [] # List to hold paths of successfully created MP3s for cleanup

    if not mp4_files:
        yield "No .mp4 files found. Skipping video conversion stage."
    else:
        yield f"Found {len(mp4_files)} MP4 file(s). Starting conversion..."
        max_workers = os.cpu_count() or 4
        
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_file = {executor.submit(convert_to_mp3, f): f for f in mp4_files}
            for future in as_completed(future_to_file):
                result = future.result()
                if isinstance(result, Path): # Success returns a Path object
                    created_audio_paths.append(result)
                    yield f"✅ Success: Converted '{future_to_file[future].name}' to '{result.name}'"
                else: # Failure returns an error string
                    yield result
    
    yield "\n--- STAGE 2 of 3: Transcribing Audio to Subtitles ---"

    # --- STAGE 2: AUDIO TO SUBTITLE TRANSCRIPTION ---
    # Combine original audio files with newly converted ones
    original_audio_files = [folder / f for f in os.listdir(folder) if f.lower().endswith(('.mp3', '.wav'))]
    # Use a set to handle potential duplicates gracefully, then sort for consistent order
    all_audio_to_process = sorted(list(set(original_audio_files + created_audio_paths)))

    if not all_audio_to_process:
        yield "No audio files found to transcribe. Pipeline finished."
        return

    total_files = len(all_audio_to_process)
    processed_count = 0
    yield f"Found {total_files} audio file(s) to transcribe. Starting process..."

    # (SRT helper function and transcription logic are identical to before)
    def fmt_time(time_sec_float):
        ms = int(round(time_sec_float * 1000.0))
        hh, ms = divmod(ms, 3600000)
        mm, ms = divmod(ms, 60000)
        ss, ms = divmod(ms, 1000)
        return f"{hh:02d}:{mm:02d}:{ss:02d},{ms:03d}"

    for i, audio_path in enumerate(all_audio_to_process):
        filename = audio_path.name
        yield f"({i+1}/{total_files}) Transcribing: {filename}"
        
        # Wrapped transcription in a try-except to ensure pipeline continues
        try:
            res_list = model.generate(input=str(audio_path), cache={}, language=language, use_itn=True, output_timestamp=True)
            # (The complex SRT generation logic is unchanged)
            srt_lines = []
            srt_entry_index = 1
            MIN_SUB_DURATION_SEC = 0.2
            
            for result_item in res_list:
                if not isinstance(result_item, dict): continue

                if "timestamp" in result_item and result_item["timestamp"]:
                    all_token_timestamps = result_item["timestamp"]
                    if not all_token_timestamps: continue
                    
                    MAX_CHARS_PER_SUB, MAX_DURATION_SEC, MIN_TOKENS_PER_SUB = 70, 7, 1
                    current_chunk_tokens_data, current_chunk_text_parts = [], []

                    for token_data in all_token_timestamps:
                        if not (len(token_data) == 3 and isinstance(token_data[0], str) and isinstance(token_data[1], (float, int)) and isinstance(token_data[2], (float, int))): continue
                        token_str, token_start_sec, token_end_sec = token_data
                        if token_end_sec < token_start_sec: token_end_sec = token_start_sec + 0.1

                        if not current_chunk_tokens_data:
                            current_chunk_tokens_data.append(token_data)
                            current_chunk_text_parts.append(token_str)
                        else:
                            prospective_text = " ".join(current_chunk_text_parts + [token_str])
                            prospective_duration = token_end_sec - current_chunk_tokens_data[0][1]
                            should_break = False
                            if token_str in '.!?。！？' and len(prospective_text) > 20:
                                should_break = True; current_chunk_tokens_data.append(token_data); current_chunk_text_parts.append(token_str)
                            elif len(prospective_text) > MAX_CHARS_PER_SUB and len(current_chunk_tokens_data) >= MIN_TOKENS_PER_SUB: should_break = True
                            elif prospective_duration > MAX_DURATION_SEC and len(current_chunk_tokens_data) >= MIN_TOKENS_PER_SUB: should_break = True

                            if should_break:
                                chunk_text_to_process, chunk_end_time = " ".join(current_chunk_text_parts), current_chunk_tokens_data[-1][2]
                                final_chunk_text, chunk_start_time = rich_transcription_postprocess(chunk_text_to_process), current_chunk_tokens_data[0][1]
                                if final_chunk_text.strip():
                                    if chunk_end_time <= chunk_start_time: chunk_end_time = chunk_start_time + MIN_SUB_DURATION_SEC
                                    srt_lines.append(f"{srt_entry_index}\n{fmt_time(chunk_start_time)} --> {fmt_time(chunk_end_time)}\n{final_chunk_text.strip()}\n\n"); srt_entry_index += 1
                                if token_str in '.!?。！？' and len(prospective_text) > 20: current_chunk_tokens_data, current_chunk_text_parts = [], []
                                else: current_chunk_tokens_data, current_chunk_text_parts = [token_data], [token_str]
                            else: current_chunk_tokens_data.append(token_data); current_chunk_text_parts.append(token_str)
                    
                    if current_chunk_tokens_data:
                        final_chunk_text, chunk_start_time, chunk_end_time = rich_transcription_postprocess(" ".join(current_chunk_text_parts)), current_chunk_tokens_data[0][1], current_chunk_tokens_data[-1][2]
                        if final_chunk_text.strip():
                            if chunk_end_time <= chunk_start_time: chunk_end_time = chunk_start_time + MIN_SUB_DURATION_SEC
                            srt_lines.append(f"{srt_entry_index}\n{fmt_time(chunk_start_time)} --> {fmt_time(chunk_end_time)}\n{final_chunk_text.strip()}\n\n")

            if not srt_lines:
                yield f"({i+1}/{total_files}) File {filename}: Could not generate any SRT entries."
            else:
                base_name = os.path.splitext(filename)[0]
                srt_path = os.path.join("sub", f"{base_name}.srt")
                with open(srt_path, "w", encoding="utf-8") as f: f.write("".join(srt_lines))
                yield f"({i+1}/{total_files}) ✅ Transcription complete: {filename} -> {srt_path}"
                processed_count += 1
        except Exception as e:
            yield f"({i+1}/{total_files}) ❌ Transcription FAILED for {filename}. Error: {e}"
            traceback.print_exc()

    # --- STAGE 3: CLEANUP ---
    if created_audio_paths:
        yield "\n--- STAGE 3 of 3: Cleaning up intermediate audio files ---"
        # Small delay to ensure files are not locked
        time.sleep(1) 
        deleted_count = 0
        for audio_file in created_audio_paths:
            try:
                os.remove(audio_file)
                deleted_count += 1
            except OSError as e:
                yield f"Warning: Could not delete intermediate file '{audio_file.name}'. Reason: {e}"
        yield f"Cleanup complete. Deleted {deleted_count} temporary audio file(s)."
    
    yield f"\n--- ✨ All Stages Complete ✨ ---\nSuccessfully transcribed {processed_count} / {total_files} audio files.\nSubtitles are in the 'sub' folder."


# --- Part 4: Gradio UI (Same as before) ---

with gr.Blocks() as demo:
    gr.Markdown("## Video to Subtitle Batch Processor (v2 with Cleanup)")
    gr.Markdown(
        "**Instructions:**\n"
        "1. Enter the absolute path to a folder containing your **MP4 videos**.\n"
        "2. The script will first convert all MP4s to MP3s (stored temporarily in an `audios` folder).\n"
        "3. Then, it will transcribe all audio to generate SRT subtitles (saved in a `sub` folder).\n"
        "4. **Finally, it will automatically delete the temporary audio files.**"
    )
    
    with gr.Row():
        folder_input = gr.Textbox(label="Video Folder Path", placeholder="e.g., C:\\Users\\YourName\\MyVideos")
        lang_input = gr.Dropdown(choices=["auto", "zh", "yue", "en", "ja", "ko"],
                                 value="ja", label="Target Language for Transcription")

    run_button = gr.Button("Start Full Pipeline (Video -> Audio -> Subtitles -> Cleanup)")
    status_output = gr.Textbox(label="Processing Log", lines=20, interactive=False, autoscroll=True)

    run_button.click(fn=video_to_subtitle_pipeline,
                     inputs=[folder_input, lang_input],
                     outputs=[status_output])

if __name__ == "__main__":
    demo.launch()