# video_to_audio.py

import os
import subprocess
import sys
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed

def check_ffmpeg():
    """Checks if ffmpeg is installed and available in the system's PATH."""
    try:
        # Run 'ffmpeg -version' and suppress output.
        # If it runs without error, ffmpeg is found.
        subprocess.run(
            ['ffmpeg', '-version'], 
            check=True, 
            capture_output=True
        )
        return True
    except (FileNotFoundError, subprocess.CalledProcessError):
        # FileNotFoundError means 'ffmpeg' command was not found.
        # CalledProcessError can happen for other reasons, but is unlikely for '-version'.
        return False

def convert_to_mp3(video_path: Path):
    """
    Converts a single video file to an MP3 audio file using ffmpeg.

    Args:
        video_path (Path): The path object of the input MP4 video.

    Returns:
        str: A message indicating the result of the conversion.
    """
    # Define the output file path by replacing the extension with .mp3
    output_path = video_path.with_suffix('.mp3')
    video_filename = video_path.name
    output_filename = output_path.name

    # Inform the user which file is being processed
    print(f"-> Starting conversion for: {video_filename}")

    # The ffmpeg command as a list of arguments
    # This is safer than a single string, especially with paths containing spaces.
    command = [
        'ffmpeg',
        '-i', str(video_path),    # Input file
        '-q:a', '0',              # Use variable bitrate for high quality (VBR)
        '-map', 'a',              # Select only the audio stream(s)
        '-y',                     # Overwrite output file if it exists
        '-hide_banner',           # Suppress printing the ffmpeg banner
        '-loglevel', 'error',     # Only show errors, not verbose processing info
        str(output_path)          # Output file
    ]

    try:
        # Execute the command
        result = subprocess.run(
            command, 
            check=True,            # Raise an exception if ffmpeg returns a non-zero exit code
            capture_output=True,   # Capture stdout and stderr
            text=True              # Decode stdout/stderr as text
        )
        return f"✅ Success: Converted '{video_filename}' to '{output_filename}'"
    except subprocess.CalledProcessError as e:
        # This occurs if ffmpeg runs but fails for some reason (e.g., corrupted file)
        error_message = e.stderr.strip()
        return f"❌ Error converting '{video_filename}': {error_message}"
    except Exception as e:
        # Catch any other unexpected exceptions
        return f"❌ An unexpected error occurred with '{video_filename}': {e}"


def main():
    """Main function to orchestrate the video to audio conversion."""
    print("--- MP4 to MP3 Batch Converter ---")

    # 1. Check for ffmpeg dependency first
    if not check_ffmpeg():
        print("\nError: ffmpeg is not found in your system's PATH.")
        print("Please download ffmpeg, extract it, and add its 'bin' folder to your PATH.")
        print("Download from: https://ffmpeg.org/download.html")
        sys.exit(1) # Exit the script with an error code

    # 2. Get folder path from user
    folder_path_str = input("Enter the absolute path of the folder containing MP4 files: ").strip()

    # 3. Validate the path
    folder_path = Path(folder_path_str)
    if not folder_path.is_dir():
        print(f"\nError: The path '{folder_path}' is not a valid directory.")
        sys.exit(1)

    # 4. Find all .mp4 files in the directory
    print(f"\nScanning for .mp4 files in '{folder_path}'...")
    mp4_files = list(folder_path.glob('*.mp4'))

    if not mp4_files:
        print("No .mp4 files were found in the specified directory.")
        return

    total_files = len(mp4_files)
    print(f"Found {total_files} MP4 file(s) to convert.")

    # 5. Set up and run the multi-threaded conversion
    # Use os.cpu_count() for a sensible number of parallel threads.
    # Default to 4 if os.cpu_count() returns None.
    max_workers = os.cpu_count() or 4
    print(f"Starting conversion with up to {max_workers} parallel threads...\n")

    completed_count = 0
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Create a future for each file conversion
        future_to_file = {executor.submit(convert_to_mp3, f): f for f in mp4_files}

        # Process results as they are completed
        for future in as_completed(future_to_file):
            result_message = future.result()
            completed_count += 1
            print(f"[{completed_count}/{total_files}] {result_message}")

    print("\n--- All tasks completed. ---")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\n\nProcess interrupted by user. Exiting.")
        sys.exit(0)