import gradio as gr
import torch
# import re # Not strictly needed for the new SRT logic if only using full segment text
import os
import contextlib # Keep for wave, though less critical now for SRT timing
import wave       # Keep for wave

from funasr import AutoModel
from funasr.utils.postprocess_utils import rich_transcription_postprocess

# --- 1. 初始化模型 (与原代码相同) ---
os.makedirs("sub", exist_ok=True)
print("正在加载模型，请稍候...")
device = "cuda:0" if torch.cuda.is_available() else "cpu"
model = AutoModel(
    model="iic/SenseVoiceSmall",
    trust_remote_code=True,
    vad_model="fsmn-vad",
    vad_kwargs={"max_single_segment_time": 30000},
    device=device,
    # Passdown kwargs for the underlying model.inference if needed by AutoModel's wrapper
    # For SenseVoiceSmall, `model_path` etc. are already part of `kwargs` in its from_pretrained
    # and AutoModel handles this.
)
print("模型加载完成。")


# --- 2. 修改核心处理函数，用于处理整个文件夹 ---
def transcribe_folder(folder_path, language):
    """
    遍历指定文件夹中的所有音频文件，进行语音识别并生成SRT字幕。
    Utilizes CTC-aligned timestamps if available from the model and chunks them.
    """
    if not folder_path or not os.path.isdir(folder_path):
        yield "错误：请输入一个有效的文件夹路径。"
        return

    audio_files = [f for f in os.listdir(folder_path) if f.lower().endswith(('.mp3', '.wav'))]
    if not audio_files:
        yield f"在文件夹 '{folder_path}' 中未找到 .mp3 或 .wav 文件。"
        return

    total_files = len(audio_files)
    processed_count = 0
    yield f"找到 {total_files} 个音频文件，开始处理..."

    def fmt_time(time_sec_float): # Accepts float seconds
        ms = int(round(time_sec_float * 1000.0)) # Convert seconds to ms and round
        hh = ms // 3600000
        mm = (ms % 3600000) // 60000
        ss = (ms % 60000) // 1000
        sub_ms = ms % 1000
        return f"{hh:02d}:{mm:02d}:{ss:02d},{sub_ms:03d}"

    for i, filename in enumerate(audio_files):
        audio_path = os.path.join(folder_path, filename)
        yield f"({i+1}/{total_files}) 正在处理: {filename}"

        try:
            res_list = model.generate(
                input=audio_path,
                cache={},
                language=language,
                use_itn=True,
                output_timestamp=True
            )
            
            if not res_list:
                is_empty = True
                # Check if any item in res_list is a dict with meaningful content
                for item_check in res_list:
                     if isinstance(item_check, dict) and \
                        (item_check.get("text", "").strip() or item_check.get("timestamp") or item_check.get("sentence_info")):
                        is_empty = False
                        break
                if is_empty:
                    yield f"({i+1}/{total_files}) 文件 {filename} 识别结果为空或无时间戳信息，已跳过。"
                    continue
            
            srt_lines = []
            srt_entry_index = 1
            
            # This loop might run once if res_list contains one aggregated dictionary,
            # or multiple times if it contains per-VAD segment dictionaries.
            # The chunking logic below handles both cases by processing the "timestamp" field.
            for result_item in res_list:
                if not isinstance(result_item, dict):
                    continue

                # OPTION 1: Prioritize detailed "timestamp" field for chunking
                if "timestamp" in result_item and result_item["timestamp"]:
                    all_token_timestamps = result_item["timestamp"] # List of [token_str, abs_start_sec, abs_end_sec]
                    
                    if not all_token_timestamps:
                        yield f"({i+1}/{total_files}) 文件 {filename}: 'timestamp' 键存在但列表为空。"
                        # continue # Don't continue the outer loop, try other fallbacks for this result_item
                    else:
                        # Chunking parameters
                        MAX_CHARS_PER_SUB = 70  # Approximate max characters
                        MAX_DURATION_SEC = 7    # Max duration in seconds
                        MIN_SUB_DURATION_SEC = 0.2 # Min duration to avoid zero/negative
                        MIN_TOKENS_PER_SUB = 1  # Ensure subtitle has at least one token before breaking

                        current_chunk_tokens_data = [] # Stores [token_str, start_sec, end_sec]
                        current_chunk_text_parts = []  # Stores just token_str for joining

                        for token_idx, token_data in enumerate(all_token_timestamps):
                            if not (len(token_data) == 3 and isinstance(token_data[0], str) and \
                                    isinstance(token_data[1], (float, int)) and \
                                    isinstance(token_data[2], (float, int))):
                                # print(f"Warning: Skipping malformed token data: {token_data}")
                                continue
                            
                            token_str, token_start_sec, token_end_sec = token_data

                            # Basic validation for token times
                            if token_end_sec < token_start_sec:
                                token_end_sec = token_start_sec + MIN_SUB_DURATION_SEC/2 # Give it a tiny duration


                            if not current_chunk_tokens_data: # First token of a new subtitle chunk
                                current_chunk_tokens_data.append(token_data)
                                current_chunk_text_parts.append(token_str)
                            else:
                                # Try adding current token to existing chunk
                                prospective_text_len = len(" ".join(current_chunk_text_parts + [token_str])) # Or "" for Japanese
                                prospective_duration = token_end_sec - current_chunk_tokens_data[0][1]

                                # Determine if we need to break and form a new subtitle
                                should_break = False
                                force_break_after_this_token = False # New flag

                                # Rule 1: Break if this token is sentence-ending punctuation AND current chunk is reasonably long
                                if token_str in ['.', '!', '?', '。', '！', '？'] and \
                                   len(" ".join(current_chunk_text_parts)) > MAX_CHARS_PER_SUB * 0.3: # Avoid breaking tiny chunks
                                    # Include this punctuation in the current subtitle, then break
                                    current_chunk_tokens_data.append(token_data)
                                    current_chunk_text_parts.append(token_str)
                                    should_break = True
                                    force_break_after_this_token = True # Indicates this token *ends* the current sub

                                # Rule 2: Break if max char limit exceeded (only if not already breaking for punctuation)
                                elif prospective_text_len > MAX_CHARS_PER_SUB and not force_break_after_this_token:
                                    if len(current_chunk_tokens_data) >= MIN_TOKENS_PER_SUB: # Ensure we don't create empty subs
                                        should_break = True

                                # Rule 3: Break if max duration limit exceeded (only if not already breaking)
                                elif prospective_duration > MAX_DURATION_SEC and not force_break_after_this_token:
                                     if len(current_chunk_tokens_data) >= MIN_TOKENS_PER_SUB:
                                        should_break = True
                                
                                if should_break:
                                    # Finalize the current chunk
                                    if force_break_after_this_token:
                                        # The current_chunk_text_parts already includes the punctuation
                                        chunk_text_to_process = " ".join(current_chunk_text_parts) # Or "" for Japanese
                                        chunk_end_time = current_chunk_tokens_data[-1][2] # End time is the punctuation's end
                                    else:
                                        # Break happened due to length/duration; current token belongs to next sub
                                        chunk_text_to_process = " ".join(current_chunk_text_parts) # Or "" for Japanese
                                        chunk_end_time = current_chunk_tokens_data[-1][2] # End time is last token of *this* sub

                                    final_chunk_text = rich_transcription_postprocess(chunk_text_to_process)
                                    chunk_start_time = current_chunk_tokens_data[0][1]
                                    
                                    if final_chunk_text.strip():
                                        if chunk_end_time <= chunk_start_time:
                                            chunk_end_time = chunk_start_time + MIN_SUB_DURATION_SEC
                                        
                                        srt_lines.append(f"{srt_entry_index}\n")
                                        srt_lines.append(f"{fmt_time(chunk_start_time)} --> {fmt_time(chunk_end_time)}\n")
                                        srt_lines.append(f"{final_chunk_text.strip()}\n\n")
                                        srt_entry_index += 1

                                    # Start new chunk
                                    if force_break_after_this_token:
                                        current_chunk_tokens_data = []
                                        current_chunk_text_parts = []
                                    else: # Current token starts the new chunk
                                        current_chunk_tokens_data = [token_data]
                                        current_chunk_text_parts = [token_str]
                                else: # No break, add token to current chunk
                                    current_chunk_tokens_data.append(token_data)
                                    current_chunk_text_parts.append(token_str)
                        
                        # Add any remaining tokens in the last chunk
                        if current_chunk_tokens_data:
                            final_chunk_text = rich_transcription_postprocess(" ".join(current_chunk_text_parts))
                            chunk_start_time = current_chunk_tokens_data[0][1]
                            chunk_end_time = current_chunk_tokens_data[-1][2]

                            if final_chunk_text.strip():
                                if chunk_end_time <= chunk_start_time:
                                    chunk_end_time = chunk_start_time + MIN_SUB_DURATION_SEC
                                srt_lines.append(f"{srt_entry_index}\n")
                                srt_lines.append(f"{fmt_time(chunk_start_time)} --> {fmt_time(chunk_end_time)}\n")
                                srt_lines.append(f"{final_chunk_text.strip()}\n\n")
                                srt_entry_index += 1
                        # After processing timestamps, we've handled this result_item, so skip other fallbacks.
                        # If srt_lines were populated, it means this path was successful.
                        if srt_lines and srt_entry_index > 1: # Check if any subs were actually added from this path
                           # This result_item is processed, move to next result_item in res_list
                           continue 
                           
                # OPTION 2: Fallback to "sentence_info" (if OPTION 1 failed or wasn't present for this result_item)
                # This assumes result_item is a per-VAD-segment dictionary
                if "sentence_info" in result_item and result_item["sentence_info"]:
                    segment_abs_start_ms = result_item.get("start_ms", 0) # VAD segment start if applicable
                    for sentence_data in result_item["sentence_info"]:
                        text_to_process = sentence_data.get("text", "")
                        processed_sentence_text = rich_transcription_postprocess(text_to_process)
                        if not processed_sentence_text.strip():
                            continue

                        # Times from sentence_data are usually relative to VAD segment
                        # If auto_model.py made them absolute, this needs adjustment.
                        # For now, assume they are relative as per typical FunASR sentence_info.
                        rel_start_ms = sentence_data.get("start", 0)
                        rel_end_ms = sentence_data.get("end", 0)
                        
                        abs_start_s = (segment_abs_start_ms + rel_start_ms) / 1000.0
                        abs_end_s = (segment_abs_start_ms + rel_end_ms) / 1000.0

                        if abs_end_s <= abs_start_s:
                             abs_end_s = abs_start_s + MIN_SUB_DURATION_SEC

                        srt_lines.append(f"{srt_entry_index}\n")
                        srt_lines.append(f"{fmt_time(abs_start_s)} --> {fmt_time(abs_end_s)}\n")
                        srt_lines.append(f"{processed_sentence_text}\n\n")
                        srt_entry_index += 1
                    if srt_lines and srt_entry_index > 1: continue # Processed via sentence_info

                # OPTION 3: Fallback to full segment text (if OPTION 1 & 2 failed for this result_item)
                # This assumes result_item is a per-VAD-segment dictionary
                segment_text_raw_fallback = result_item.get("text", "").strip()
                if segment_text_raw_fallback: # If there's text but no fine-grained timestamps
                    processed_text_fallback = rich_transcription_postprocess(segment_text_raw_fallback)
                    if processed_text_fallback.strip():
                        start_ms_fallback = result_item.get("start_ms", 0)
                        end_ms_fallback = result_item.get("end_ms") 
                        length_ms_fallback = result_item.get("length_ms")

                        if end_ms_fallback is None and length_ms_fallback is not None:
                            end_ms_fallback = start_ms_fallback + length_ms_fallback
                        
                        if end_ms_fallback is not None and end_ms_fallback > start_ms_fallback:
                            srt_lines.append(f"{srt_entry_index}\n")
                            srt_lines.append(f"{fmt_time(start_ms_fallback / 1000.0)} --> {fmt_time(end_ms_fallback / 1000.0)}\n")
                            srt_lines.append(f"{processed_text_fallback}\n\n")
                            srt_entry_index += 1
                        # else:
                        #     yield f"({i+1}/{total_files}) 文件 {filename}, 片段 fallback: 无有效时长。"


            if not srt_lines:
                yield f"({i+1}/{total_files}) 文件 {filename}: 未能生成任何SRT条目。"
                # continue # This continue is for the outer loop over audio_files
            else: # Only save if srt_lines were generated
                base_name = os.path.splitext(filename)[0]
                srt_path = os.path.join("sub", f"{base_name}.srt")
                with open(srt_path, "w", encoding="utf-8") as f:
                    f.write("".join(srt_lines))
                
                yield f"({i+1}/{total_files}) ✅ 处理完成: {filename} -> 字幕已保存至 {srt_path}"
                processed_count += 1
            
        except Exception as e:
            yield f"({i+1}/{total_files}) ❌ 处理失败: {filename}. 错误: {e}"
            import traceback
            traceback.print_exc() 

    yield f"\n--- ✨ 全部处理完成 ✨ ---\n成功处理 {processed_count} / {total_files} 个文件。\n所有字幕文件均已保存在 'sub' 文件夹中。"


# --- 3. 创建新的 Gradio 界面 (与原代码相同) ---
with gr.Blocks() as demo:
    gr.Markdown("## SenseVoiceSmall 批量语音识别与字幕生成 (CTC 精确时间戳)")
    gr.Markdown("请在下方输入存放 MP3/WAV 音频文件的**文件夹路径**，然后点击开始处理。字幕将优先使用模型提供的CTC对齐时间戳。")
    
    with gr.Row():
        folder_input = gr.Textbox(label="音频文件夹路径", placeholder="例如: /home/user/my_audios 或 C:\\Users\\User\\audios")
        lang_input = gr.Dropdown(choices=["auto", "zh", "yue", "en", "ja", "ko"],
                                 value="ja", label="识别语言")

    run_button = gr.Button("开始批量处理")
    status_output = gr.Textbox(label="处理状态", lines=15, interactive=False, autoscroll=True) # Added autoscroll

    run_button.click(fn=transcribe_folder,
                     inputs=[folder_input, lang_input],
                     outputs=[status_output])

if __name__ == "__main__":
    demo.launch()